package orm

import (
	"context"
	"fmt"
	"github.com/stretchr/testify/assert"
	"strings"
	"testing"
	"time"

	"github.com/go-loremipsum/loremipsum"
)

type chapter struct {
	ID            bson.ObjectID `json:"_id"`
	Title         string        `json:"chapterTitle" form:"chapterTitle"`
	ChapterID     int           `json:"chapterID" autoinc:"chapters"`
	Index         int           `json:"index" form:"index"`
	Words         int           `json:"words"`
	Notes         string        `json:"notes" form:"notes"`
	Genre         []string      `json:"genre" form:"genre"`
	Bands         []band        `json:"bands" ref:"band,bands"`
	Characters    []string      `json:"characters" form:"characters"`
	Relationships [][]string    `json:"relationships" form:"relationships"`
	Adult         bool          `json:"adult" form:"adult"`
	Summary       string        `json:"summary" form:"summary"`
	Hidden        bool          `json:"hidden" form:"hidden"`
	LoggedInOnly  bool          `json:"loggedInOnly" form:"loggedInOnly"`
	Posted        time.Time     `json:"datePosted"`
	FileName      string        `json:"fileName"`
	Text          string        `json:"text" gridfs:"story_text,/stories/{{.ChapterID}}.txt"`
}

type band struct {
	ID         int64 `json:"_id"`
	Document   `json:",inline" coll:"bands"`
	Name       string   `json:"name" form:"name"`
	Locked     bool     `json:"locked" form:"locked"`
	Characters []string `json:"characters" form:"characters"`
}
type user struct {
	ID       int64 `json:"_id"`
	Document `json:",inline" coll:"users"`
	Username string `json:"username"`
	Favs     []user `json:"favs" ref:"user"`
}
type story struct {
	ID        int64 `json:"_id"`
	Document  `json:",inline" coll:"stories"`
	Title     string    `json:"title" form:"title"`
	Author    *user     `json:"author" ref:"user"`
	CoAuthor  *user     `json:"coAuthor" ref:"user"`
	Chapters  []chapter `json:"chapters"`
	Recs      int       `json:"recs"`
	Favs      int       `json:"favs"`
	Views     int       `json:"views"`
	Completed bool      `json:"completed" form:"completed"`
	Downloads int       `json:"downloads"`
}
type somethingWithNestedChapters struct {
	ID         int64 `json:"_id"`
	Document   `json:",inline" coll:"nested_stuff"`
	Chapters   []chapter `json:"chapters"`
	NestedText string    `json:"text" gridfs:"nested_text,/nested/{{.ID}}.txt"`
}

func (s *somethingWithNestedChapters) Id() any {
	return s.ID
}

func (s *somethingWithNestedChapters) SetId(id any) {
	s.ID = id.(int64)
}

func (s *story) Id() any {
	return s.ID
}

func (s *band) Id() any {
	return s.ID
}
func (s *user) Id() any {
	return s.ID
}

func (s *story) SetId(id any) {
	s.ID = id.(int64)
	//var t IDocument =s
}

func (s *band) SetId(id any) {
	s.ID = id.(int64)
}

func (s *user) SetId(id any) {
	s.ID = id.(int64)
}

var author = user{
	Username: "tablet.exe",
	Favs: []user{
		{
			Username: "DarQuiel7",
		},
	},
}

func genChaps(single bool) []chapter {
	var ret []chapter
	var ceil int
	if single {
		ceil = 1
	} else {
		ceil = 5
	}
	emptyRel := make([][]string, 0)
	emptyRel = append(emptyRel, make([]string, 0))
	relMap := [][][]string{
		{
			{"Sean Harris", "Brian Tatler"},
		},
		{
			{"Sean Harris", "Brian Tatler"},
			{"Duncan Scott", "Colin Kimberley"},
		},
		{
			{"Duncan Scott", "Colin Kimberley"},
		},
		emptyRel,
		{
			{"Sean Harris", "Colin Kimberley", "Brian Tatler"},
		},
	}
	l := loremipsum.New()

	for i := 0; i < ceil; i++ {
		spf := fmt.Sprintf("%d.md", i+1)
		ret = append(ret, chapter{
			ID:            bson.NewObjectID(),
			Title:         fmt.Sprintf("-%d-", i+1),
			Index:         i + 1,
			Words:         50,
			Notes:         "notenotenote !!!",
			Genre:         []string{"Slash"},
			Bands:         []band{diamondHead},
			Characters:    []string{"Sean Harris", "Brian Tatler", "Duncan Scott", "Colin Kimberley"},
			Relationships: relMap[i],
			Adult:         true,
			Summary:       l.Paragraph(),
			Hidden:        false,
			LoggedInOnly:  true,
			FileName:      spf,
			Text:          strings.Join(l.ParagraphList(10), "\n\n"),
		})
	}

	return ret
}

func doSomethingWithNested() somethingWithNestedChapters {
	l := loremipsum.New()
	swnc := somethingWithNestedChapters{
		Chapters:   genChaps(false),
		NestedText: strings.Join(l.ParagraphList(15), "\n\n"),
	}
	return swnc
}
func iti_single() story {
	return story{
		Title:     "title",
		Completed: true,
		Chapters:  genChaps(true),
	}
}

func iti_multi() story {
	return story{
		Title:     "Brian Tatler Fucked and Abused Sean Harris",
		Completed: false,
		Chapters:  genChaps(false),
	}
}

func iti_blank() story {
	t := iti_single()
	t.Chapters = make([]chapter, 0)
	return t
}

func initTest() {
	uri := "mongodb://127.0.0.1:27017"
	db := "rockfic_ormTest"
	ic, _ := mongo.Connect(options.Client().ApplyURI(uri))
	ic.Database(db).Drop(context.TODO())
	colls, _ := ic.Database(db).ListCollectionNames(context.TODO(), bson.M{})
	if len(colls) < 1 {
		mdb := ic.Database(db)
		mdb.CreateCollection(context.TODO(), "bands")
		mdb.CreateCollection(context.TODO(), "stories")
		mdb.CreateCollection(context.TODO(), "users")
	}
	defer ic.Disconnect(context.TODO())
	Connect(uri, db)
	author.ID = 696969
	ModelRegistry.Model(band{}, user{}, story{})
}

var metallica = band{
	ID:   1,
	Name: "Metallica",
	Characters: []string{
		"James Hetfield",
		"Lars Ulrich",
		"Kirk Hammett",
		"Cliff Burton",
	},
	Locked: false,
}

var diamondHead = band{
	ID:     503,
	Name:   "Diamond Head",
	Locked: false,
	Characters: []string{
		"Brian Tatler",
		"Sean Harris",
		"Duncan Scott",
		"Colin Kimberley",
	},
}
var bodom = band{
	ID:     74,
	Name:   "Children of Bodom",
	Locked: false,
	Characters: []string{
		"Janne Wirman",
		"Alexi Laiho",
		"Jaska Raatikainen",
		"Henkka T. Blacksmith",
		"Roope Latvala",
		"Daniel Freyberg",
		"Alexander Kuoppala",
	},
}

func saveDoc(t *testing.T, doc IDocument) {
	err := doc.Save()
	assert.Nil(t, err)
}

func createAndSave(t *testing.T, doc IDocument) {
	mdl := Create(doc).(IDocument)
	saveDoc(t, mdl)
}
